







      subroutine pdcrhs( av1,    av3,    b,      bound,  bposit, c,
     +                   endn,   larcs,  lfirst, lsets,  miu,    na,
     +                   nbound, nn,     npfrst, rhs,    root,   rsets, 
     +                   s,      strn,   teta,   u,      uarcs,  ufirst,
     +                   upfrst, upnext, w,      wd,     x,      xd,
     +                   y,      z,      zd )
c     ------------------------------------------------------------------
c     pdcrhs : This subroutine computes the right-hand side of the 
c              normal equations system to be solved in the network 
c              primal-dual interior-point algorithm.  This subroutine 
c              predefines the sets l and u.
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             bound : number of bounded arcs.
c             larcs : number of arcs at the lower bound.
c             lfirst: first arc at the lower bound.
c             na    : number of arcs.
c             nn    : number of nodes.
c             nbound: number of unbounded arcs.
c             npfrst: next bounded arc to be considered.
c             root  : sets beginning of structure.
c             uarcs : number of arcs at the upper bound.
c             ufirst: first arc at the upper bound.
c             upfrst: next unbounded arc to be considered.
c     ------------------------------------------------------------------
      integer bound , larcs , lfirst, na    , nn    , nbound,
     +        npfrst, root  , uarcs , ufirst, upfrst

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c          bposit(1:na): flags arcs according to bounds.
c          endn  (1:na): pointer of network data structure.
c          lsets (1:na): list of arcs at lower bound.
c          rsets (1:na): list of arcs at upper bound.
c          strn  (1:na): adjacency list of network data structure.
c          upnext(1:na): next arc with positive upper bound.
c     ------------------------------------------------------------------
      integer
     +     bposit(1:na),
     +     endn  (1:na),
     +     lsets (1:na),
     +     rsets (1:na),
     +     strn  (1:na),
     +     upnext(1:na)

c     ------------------------------------------------------------------
c     Double precision input/output arrays:
c
c          av1 (1:na): auxilliary vector 1.
c          av3 (1:na): auxilliary vector 3.
c          b   (1:nn): array of capacities at each node.
c          c   (1:na): array of costs at each arc.
c          rhs (1:nn): right-hand side of the linear system.
c          s   (1:na): dual iterate s.
c          teta(1:na): normal equations diagonal matrix.
c          u   (1:na): upper bound of flow at each arc.
c          w   (1:na): dual iterate w.
c          wd  (1:na): direction dw.
c          x   (1:na): primal iterate x.
c          xd  (1:na): direction dx.
c          y   (1:nn): primal iterate y.
c          z   (1:na): primal iterate z.
c          zd  (1:na): direction dz.
c     ------------------------------------------------------------------
      double precision
     +     av1 (1:na),
     +     av3 (1:na),
     +     b   (1:nn),
     +     c   (1:na),
     +     rhs (1:nn),
     +     s   (1:na),
     +     teta(1:na),
     +     u   (1:na),
     +     w   (1:na),
     +     wd  (1:na),
     +     x   (1:na),
     +     xd  (1:na),
     +     y   (1:nn),
     +     z   (1:na),
     +     zd  (1:na)

c     ------------------------------------------------------------------
c     Integer variables.
c
c             arc   : current arc.
c             enod  : current element on endn.
c             i     : loop counter.
c             node  : another loop counter.
c             luxarc: current arc on lower bound.
c             snod  : current element on strn.
c             uuxarc: current arc on upper bound.
c     ------------------------------------------------------------------
      integer arc   , enod  , i     , node  , luxarc, snod  , uuxarc

c     ------------------------------------------------------------------
c     Double precision input/output variables:
c
c             miu: parameter mu.
c     ------------------------------------------------------------------
      double precision miu

c
c     Double precision variables.
c
c            aux1-7: auxilliary variables.
c     ------------------------------------------------------------------
      double precision  aux1  , aux2  , aux3  , aux4  ,
     +                  aux5  , aux6  , aux7

c     ------------------------------------------------------------------
c     Start of executable section of subroutine.
c     
c     Setting lfirst and ufirst equal to zero.
c     ------------------------------------------------------------------
      lfirst = 0
      ufirst = 0
      larcs  = 0
      uarcs  = 0

c     ------------------------------------------------------------------
c     Initialize the right-hand side vector rhs.
c     ------------------------------------------------------------------
      do 10 node=1,root-1
         rhs(node) = b(node)
 10   continue
      do 20 node=root+1,nn
         rhs(node) = b(node)
 20   continue
      rhs(root) = 0d0

c     ------------------------------------------------------------------
c     Compute the right-hand side for pcg.
c     Define sets b, l and u.
c     
c     
c     Arcs without upper bound.
c     ------------------------------------------------------------------
      arc = npfrst
      do 30 i=1,nbound
c        ---------------------------------------------------------------
c        Compute working parameters.
c        ---------------------------------------------------------------
         snod = strn(arc)
         enod = endn(arc)
         aux1 = z(arc)/x(arc)
         aux3 = miu/x(arc)
         aux6 = (aux3-y(snod)+y(enod)-c(arc))/aux1
         aux7 = aux6+x(arc)

c        ---------------------------------------------------------------
c        Compute the right-hand side for pcg.
c        ---------------------------------------------------------------
         rhs(snod) = rhs(snod)+aux7
         rhs(enod) = rhs(enod)-aux7

c        ---------------------------------------------------------------
c        Predefine set l.
c        ---------------------------------------------------------------
         if (lfirst.eq.0) then
            lfirst         = arc
            lsets(lfirst)  = 0
            rsets(lfirst)  = 0
            luxarc        = lfirst
         else
            lsets(arc)     = luxarc
            rsets(arc)     = 0
            rsets(luxarc) = arc
            luxarc        = arc
         endif
         bposit(arc) = -1
         larcs = larcs+1

c        ---------------------------------------------------------------
c        Compute auxilliary vectors.
c        ---------------------------------------------------------------
         xd(arc)   = aux6
         zd(arc)   = aux3
         av1(arc)  = aux1
         teta(arc) = aux1

c        ---------------------------------------------------------------
c        Compute the next arc.
c        ---------------------------------------------------------------
         arc = upnext(arc)
 30   continue

c     ------------------------------------------------------------------
c     Arcs with upper bound
c     ------------------------------------------------------------------
      arc = upfrst
      do 40 i=1,bound
c        ---------------------------------------------------------------
c        Compute working parameters.
c        ---------------------------------------------------------------
         snod = strn(arc)
         enod = endn(arc)
         aux1 = z(arc)/x(arc)
         aux2 = w(arc)/s(arc)
         aux3 = miu/x(arc)
         aux4 = miu/s(arc)
         aux5 = aux1+aux2
         aux6 = (aux3-aux4-y(snod)+y(enod)-c(arc)+
     +           aux2*(u(arc)-x(arc))-w(arc))/aux5
         aux7 = aux6+x(arc)

c        ---------------------------------------------------------------
c        Compute the right-hand side for pcg.
c        ---------------------------------------------------------------
         rhs(snod) = rhs(snod)+aux7
         rhs(enod) = rhs(enod)-aux7

c        ---------------------------------------------------------------
c        Predefine sets l and u.
c        ---------------------------------------------------------------
         if (aux1.ge.aux2) then
            if (lfirst.eq.0) then
               lfirst         = arc
               lsets(lfirst)  = 0
               rsets(lfirst)  = 0
               luxarc         = lfirst
            else
               lsets(arc)     = luxarc
               rsets(arc)     = 0
               rsets(luxarc) = arc
               luxarc        = arc
            endif
            larcs             = larcs+1
            bposit(arc)       = -1
         else
            if (ufirst.eq.0) then
               ufirst         = arc
               lsets(ufirst)  = 0
               rsets(ufirst)  = 0
               uuxarc        = ufirst
            else
               lsets(arc)     = uuxarc
               rsets(arc)     = 0
               rsets(uuxarc) = arc
               uuxarc        = arc
            endif
            uarcs             = uarcs+1
            bposit(arc)       = 1
         endif

c        ---------------------------------------------------------------
c        Compute auxilliary vectors.
c        ---------------------------------------------------------------
         xd(arc)   = aux6
         zd(arc)   = aux3
         wd(arc)   = aux4
         av1(arc)  = aux1
         av3(arc)  = aux2
         teta(arc) = aux5

c        ---------------------------------------------------------------
c        Compute the next arc.
c        ---------------------------------------------------------------
         arc = upnext(arc)
 40   continue

c     ------------------------------------------------------------------
c     Setting rhs(root) and av2(root) equal to zero.
c     ------------------------------------------------------------------
      rhs(root) = 0d0
      return

c     ------------------------------------------------------------------
c     End of subroutine pdcrhs.
c     ------------------------------------------------------------------
      end
