






      subroutine pdmxst( endv,  nord,  ntree,  nvrtx,  ordedg,  parent,
     +                   rank, stedg,   strv )
c     ------------------------------------------------------------------
c     pdmxst: Given a graph g and a ordering of the edges, this
c             subroutine applies Kruskal's algorithm and builds a
c             maximum spanning forest.
c
c             An ordered list with "nord" indices of allowable edges is
c             given in the input array "ordedg".  Edges are identified
c             by their start and end vertices in arrays strv and  endv.
c
c             Output is expressed by 3 items:
c
c                   1. ntree:  number of trees in spanning forest.
c                   2. stedg:  list of (nvrtx - ntree) edges in spanning
c                              forest.
c                   3. parent: tree index for each vertex.
c
c             Arrays "stedg" and "ordedg" can point to the same memory 
c             location.
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             nord  : number of edges in ordered list.
c             ntree : number of trees in spanning forest.
c             nvrtx : number of vertices.
c     ------------------------------------------------------------------
      integer nord  , ntree , nvrtx

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c             endv  (1:  nedge): end vertex in network data structure
c             ordedg(1:  nord ): edge ordering.
c             parent(1:  nvrtx): parent array for tree representation.
c             rank  (1:  nvrtx): upper bound on height of element.
c             stedg (1:  nord) : edges in spanning forest.
c             strv  (1:  nedge): start vertex in network data structure.
c     ------------------------------------------------------------------
      integer
     +        endv  (1:  nord ),
     +        ordedg(1:  nord ),
     +        parent(1:  nvrtx),
     +        rank  (1:  nvrtx),
     +        stedg (1:  nord),
     +        strv  (1:  nord )

c     ------------------------------------------------------------------
c     Integer variables:
c
c             e     : edge index.
c             ntedg : count of edges in spanning forest.
c             ord   : ordered edge list index.
c             partv : canonical element for set containing v.
c             partw : canonical element for set containing w.
c             v     : vertex index.
c             vrtx  : vertex index.
c             w     : vertex index.
c     ------------------------------------------------------------------
      integer e     , ntedg , ord   , partv , partw , v     , vrtx  , w

c     ------------------------------------------------------------------
c     Start of executable portion of subroutine.
c
c     Make sets -- initialize parent and rank.
c     ------------------------------------------------------------------
      do 10 vrtx=1,nvrtx
         parent(vrtx)=vrtx
         rank(vrtx)=0
 10   continue

c     ------------------------------------------------------------------
c     Initialize number of edges in spanning tree (forest).
c     ------------------------------------------------------------------
      ntedg=0
      do 20 ord=1,nord
c        ---------------------------------------------------------------
c        Identify next edge.
c        ---------------------------------------------------------------
         e=ordedg(ord)
         v=strv(e)
         w=endv(e)
 
c        ---------------------------------------------------------------
c        If v and w are from different partitions join both partitions.
c        ---------------------------------------------------------------
         call pdfdst( partv,  parent, v )
         call pdfdst( partw,  parent, w )
         if(partv.ne.partw) then
            call pdlkst( partv,  partw,  parent, rank )

c           ------------------------------------------------------------
c           Add e=(v,w) to max weighted spanning tree t.
c           ------------------------------------------------------------
            ntedg=ntedg+1
            stedg(ntedg)=e

c           ------------------------------------------------------------
c           Break if tree is complete.
c           ------------------------------------------------------------
            if(ntedg.eq.nvrtx-1) go to 30
         endif
 20   continue

c     ------------------------------------------------------------------
c     Initialize number of trees in spanning forest.
c     ------------------------------------------------------------------
 30   ntree=0

c     ------------------------------------------------------------------
c     Associate vertices with roots, count vertices for each root.
c     ------------------------------------------------------------------
      do 40 vrtx=1,nvrtx
         call pdfdst( v,      parent, vrtx )
         if(v.eq.vrtx) then
            ntree=ntree+1
            rank(vrtx)=ntree
         endif
 40   continue

c     ------------------------------------------------------------------
c     Translate entries in parent array and count.
c     ------------------------------------------------------------------
      do 50 vrtx=1,nvrtx
         parent(vrtx)=rank(parent(vrtx))
 50   continue
      return

c     ------------------------------------------------------------------
c     End of subroutine pdmxst.
c     ------------------------------------------------------------------
      end
