      integer function pdnet(      b,      c,   dinfo, endn,   info,    
     +     l, maxa, maxn, maxwkd,     na,    nn, nwkd, 
     +     optflo, ptrwkd, ptrwki,   strn,  u )
c     ------------------------------------------------------------------
c     pdnet: This is the main subroutine for the PDNET package.
c
c            The subroutine starts by assigning working vectors to
c            pointers, with pdpntr(). The control parameters are then
c            read from the vector info, with pdginf(). Correctness of
c            the input is checked with pdchkp(), and data structure
c            created with pdmked(). Internal parameters for methods
c            are set with pddflt(), and additional structures are
c            created with pddstr(). Subroutines pdtrfm() and
c            pdprtb() are called in order to shift the lower bounds to
c            zero and to transform the data into double precision,
c            respectively. Subroutines pdpdat() and pdckfs() check the
c            problem characteristics and inquire if the network has
c            enough capacity to transport the proposed amount of
c            commodity. The primal-dual main loop is then started and
c            the current iteration direction system right-hand-side is
c            computed by pcrhs(). The maximum spanning tree is computed
c            by pdhprm(), and its optimality is tested by pdoptm().
c            Under certain conditions (miu < 1.0d0), the maxflow
c            stopping criterion is invoked by subroutine pdchmf().
c            The preconditioner is computed next, and if iqrd=-2,
c            subroutine pdiqrd() is called to compute the IRQD
c            preconditioner. The preconditioner switch comes next,
c            and pdpcgd() is called afterwards, so that the linear
c            system is solved. Preconditioner update is then made, and
c            a summary of the iteration is printed by pdsout(). Primal
c            and dual updates are made by pdupsl() and stopping criteria
c            check takes place, before returning to the start of the
c            iteration loop.
c
c     Authors: Luis Portugal (Univ. Coimbra - Portugal)
c              Mauricio G. C. Resende (AT&T Labs Research - USA)
c              Geraldo Veiga (AT&T Labs - USA)
c              Joaquim Judice (Univ. Coimbra - Portugal)
c              Joao Patricio (EST Tomar - Portugal)
c
c     Reference: L.F. Portugal, M.G.C. Resende, G. Veiga, and J.J. Judice,
c     "A truncated primal-infeasible dual-feasible interior point network
c     flow method," Networks, vol. 35, pp. 91-108, 2000.
c
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             maxa  : maximum number of arcs.
c             maxn  : maximum number of nodes.
c             maxwkd: maximum size of double precision work array.
c             na    : number of arcs.
c             nn    : number of nodes.
c             nwkd : size of double precision work array
c     ------------------------------------------------------------------
      integer  maxa, maxn, maxwkd, na, nn, nwkd
      integer i

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c              b     (1:  nn): array of capacities at each node.
c              c     (1:  na): array of costs at each arc.
c              endn  (1:  na): pointer of network data structure.
c              info  (1:  30): array with integer statistics and
c                              parameters.
c              l     (1:  na): lower bound of flow at each arc.
c              optflo(1:  na): array with optimal flow at each arc.
c              ptrwki(1: nwkd): alias pointer to work space.
c              strn  (1:  na): adjacency list of network data structure.
c              u     (1:  na): upper bound of flow at each arc.
c     ------------------------------------------------------------------
      integer  b     (1:  nn),
     +         c     (1:  na),
     +         endn  (1:  na),
     +         info  (1:  30),
     +         l     (1:  na),
     +         optflo(1:  na),
     +         ptrwki(1:nwkd),
     +         strn  (1:  na),
     +         u     (1:  na)

c     ------------------------------------------------------------------
c     Double precision input/output arrays:
c
c             dinfo (1:    35): array with double precision statistics 
c                               and parameters.
c             ptrwkd(1:  nwkd): pointer to work space.
c     ------------------------------------------------------------------
      double precision
     +        dinfo (1:    35),
     +        ptrwkd(1:  nwkd)

c     ------------------------------------------------------------------
c     Integer working variables:
c
c              bfirst: used for building maximum spanning tree.
c              bound : used for computing aditional data structures.
c              cgstop: flags the way PCG stopped.
c              ierror: flags errors in PDNET.
c              ioflag: controls the level of output of summary.
c              iout  : output file unit.
c              iprcnd: preconditioner used.
c              larcs : auxiliary variable.
c              lfirst: auxiliary variable.
c              maxit : maximum number of primal-dual iterations.
c              mfstat: maxflow status.
c              mr    : estimate of maximum number of iterations.
c              mxcgit: maximum number of PCG iterations.
c              mxfe  : edges on optimal flow network.
c              mxfv  : vertices on optimal flow network.
c              nbound: used for computing aditional data structures.
c              nbuk  : number of buckets.
c              npfrst: First bounded arc in upnext list
c              opt   : flags optimal flow.
c              optgap: Duality gap optimality indicator flag.
c              optmf : Maximum flow optimality indicator flag.
c              optst : Spanning tree optimality indicator flag.
c              pcgit : number of PCG iterations performed.
c              pduit : number of primal-dual iterations performed.
c              pf    : iteration level.
c              prttyp: printing level.
c              rc    : flags optimality status.
c              root  : sets beginning of structure.
c              uarcs : number of arcs at the upper bound.
c              ufirst: auxiliary variable.
c              upfrst: First unbounded arc in upnext list
c     ------------------------------------------------------------------
      integer  bfirst, bound , cgstop, ierror, ioflag, iout  ,
     +         iprcnd, larcs , lfirst, maxit,
     +         mfstat, mr    , mxcgit, mxfe  , mxfv  , nbound,
     +         nbuk  , npfrst, opt   , optgap, optmf , optst , 
     +         pcgit , pduit , pf    , prttyp, rc    , root  , 
     +         uarcs , ufirst, upfrst

c    ------------------------------------------------------------------
c    Integer vectors pointers:
c
c            pibckt: pointer for bucket.
c            pibpst: pointer for bposit.
c            pibsts: pointer for bstats.
c            pifrst: pointer for first.
c            pifthr: pointer for father.
c            piifst: pointer for ifirst.
c            piinxt: pointer for innext.
c            pilist: pointer for list.
c            pillnk: pointer for link.
c            pilost: pointer for lost.
c            pilsts: pointer for lsets.
c            pimark: pointer for mark.
c            piofst: pointer for oufrst.
c            pionxt: pointer for ounext.
c            pipred: pointer for pred.
c            pirank: pointer for rank.
c            pirlnk: pointer for rlink.
c            pirsts: pointer for rsets.
c            pirtth: pointer for rtrth.
c            pisptr: pointer for stptr.
c            pistdg: pointer for stedg.
c            pistrn: pointer for strn.
c            pisvtx: pointer for stvrtx.
c            pitrrc: pointer for trearc.
c            pitrth: pointer for treath.
c            piunxt: pointer for unext.
c            piw2e1: pointer for wk2e1.
c            piwke1: pointer for wrke1.
c            piwke2: pointer for wrke2.
c            piwke3: pointer for wrke3.
c            piwke4: pointer for wrke3.
c     ------------------------------------------------------------------
      integer pibckt, pibpst, pibsts, pifrst, pifthr, piifst,
     +        piinxt, pilist, pillnk, pilost, pilsts, pimark, piofst,
     +        pionxt, pipred, pirank, pirlnk, pirsts, pirtth, pisptr,
     +        pistdg, pistrn, pisvtx, pitrrc, pitrth, piunxt, piw2e1,
     +        piwke1, piwke2, piwke3, piwke4

c     ------------------------------------------------------------------
c     Double precision vectors pointers:
c
c             prav01: pointer for av1.
c             prav03: pointer for av3.
c             prav04: pointer for av4.
c             prav05: pointer for av5.
c             prav06: pointer for av6.
c             prav07: pointer for av7.
c             prav08: pointer for av8.
c             prdb00: pointer for db.
c             prdc00: pointer for dc.
c             prdiag: pointer for diag.
c             prdu00: pointer for du.
c             prkey0: pointer for key.
c             prndig: pointer for ndiag.
c             prrhs0: pointer for rhs.
c             prteta: pointer for teta.
c             prvcsd: pointer for sd.
c             prvcwd: pointer for wd.
c             prvcxd: pointer for xd.
c             prvcyd: pointer for yd.
c             prvczd: pointer for zd.
c             prvczz: pointer for zz.
c             prvecp: pointer for p.
c             prvecq: pointer for q.
c             prvecr: pointer for r.
c             prvecs: pointer for s.
c             prvecw: pointer for w.
c             prvecx: pointer for x.
c             prvecy: pointer for y.
c             prvecz: pointer for z.
c             prwrk1: pointer for wrkdv1.
c     ----------------------------------------------------------------
      integer prav01, prav03, prav04, prav05, prav06, prav07, prav08,
     +        prdb00, prdc00, prdiag, prdu00, prkey0, prndig, prrhs0,
     +        prteta, prvcsd, prvcwd, prvcxd, prvcyd, prvczd, prvczz,
     +        prvecp, prvecq, prvecr, prvecs, prvecw, prvecx, prvecy,
     +        prvecz, prwrk1

c     ----------------------------------------------------------------
c     Double precision variables:
c
c             dobj  : dual objective function value.
c             dof   : dual objective function value on tree.
c             ds    : updating parameter for vector s.
c             factor: factor for Newton step.
c             fadd  : auxiliary variable for cost computation.
c             fcttol: factorization zero tolerance
c             gap   : value of the gap.
c             grho0 : factor for rho0 update.
c             gtolmf: update factor for tol1 and tol2.
c             huge  : maxreal number.
c             maxb  : largest absolute value of b.
c             maxc  : largest absolute value of c.
c             maxu  : largest absolute value of u.
c             mfdopt: maxflow value.
c             minb  : smallest absolute value of b.
c             minc  : smallest absolute value of c.
c             minu  : smallest absolute value of u.
c             miu   : value of interior-point mu parameter.
c             mntlcg: lower bound for CG tolerance.
c             mrho0 : lower bound for rho0.
c             norb  : 1-norm of vector b.
c             norc  : 1-norm of vector c.
c             noru  : 1-norm of vector u.
c             oldcos: value of cosine on previous PCG iteration.
c             oldtol: value of residual on previous PCG iteration.
c             pcos  : value of cosine on PCG iteration.
c             pcgres: value of residual on PCG iteration.
c             pobj  : objective function value.
c             pof   : primal objective function value on tree.
c             ps    : maximum primal stepsize.
c             rgap  : relative primal-dual gap.
c             rho0  : parameter from IQRD preconditioner.
c             s1fctr: factor for miu.
c             stptol: zero for dual slacks.
c             stpval: largest slack on dual face.
c             sw1   : iteration limit on CG for preconditioner switch.
c             sw2   : another limit for switch.
c             tol1  : lower bound of the tolerance for maxflow.
c             tol2  : upper bound of the tolerance for maxflow.
c             tolcg : tolerance for PCG stopping criterion.
c             tolcg0: guess for tolerance for PCG stopping criterion.
c             tolpcg: initial value for tolcg.
c             weight: tree weight.
c             zero  : value for zero.
c     ------------------------------------------------------------------
      double precision dobj  , dof   , ds    , factor, fadd  ,
     +                 fcttol, gap   , grho0 , gtolcg, gtolmf, huge  ,
     +                 maxb  , maxc  , maxu  , mfdopt, minb  , minc  ,
     +                 minu  , miu   , mntlcg, mrho0 , norb  , norc  ,
     +                 noru  , oldcos, oldtol, pcgcos, pcgres, pobj  ,
     +                 pof   , ps    , rgap  , rho0  , s1fctr, stptol,
     +                 stpval, sw1   , sw2   , tol1  , tol2  , tolcg ,
     +                 tolcg0, tolpcg, tolslk, tolsw1, tolsw2, weight,
     +                 zero

c     ------------------------------------------------------------------
c     Start of executable section of the subroutine.
c     ------------------------------------------------------------------
      ierror = 0


c     ------------------------------------------------------------------
c     Assign pointers for work arrays.
c     ------------------------------------------------------------------
      call pdpntr( na,     nn,      nwkd,  pibckt, pibpst, pibsts,
     +             pifrst, pifthr, piifst, piinxt, pilist, pillnk,
     +             pilost, pilsts, pimark, piofst, pionxt, pipred,
     +             pirank, pirlnk, pirsts, pirtth, pisptr, pistdg,
     +             pistrn, pisvtx, pitrrc, pitrth, piunxt, piw2e1,
     +             piwke1, piwke2, piwke3, piwke4, prav01, prav03,
     +             prav04, prav05, prav06, prav07, prav08, prdb00,
     +             prdc00, prdiag, prdu00, prkey0, prndig, prrhs0,
     +             prteta, prvcsd, prvcwd, prvcxd, prvcyd,
     +             prvczd, prvczz, prvecp, prvecq, prvecr, prvecs,
     +             prvecw, prvecx, prvecy, prvecz, prwrk1
     +             )

     
cc     ------------------------------------------------------------------
cc     Get imported parameter values from info array
cc     ------------------------------------------------------------------
c      call pdginf( bound , cgstop, dinfo , dobj  , dof   , factor,
c     +             fcttol, gap   , grho0 , gtolcg, gtolmf, huge  ,
c     +             ierror, info  , ioflag, iout  , iprcnd,
c     +             maxit , mfdopt, mfstat, miu   , mntlcg, 
c     +             mrho0 , mxcgit, mxfe  , mxfv  , nbound, nbuk  ,
c     +             oldcos, oldtol, opt   , optgap, optmf , optst ,
c     +             pcgcos, pcgit , pcgres, pduit , pf    , pobj  , 
c     +             pof   , prttyp, rho0  , root  , s1fctr, stptol, 
c     +             stpval, tol1  , tol2  , tolcg , tolcg0, tolpcg, 
c     +             tolslk, zero)
 
c     ------------------------------------------------------------------
c     Set default values for internal parameters
c     ------------------------------------------------------------------     
      call pddflt(dinfo, info, na, nn)


      call pdginf( bound,  cgstop, dinfo,  dobj,   dof,    factor,
     +             fcttol, gap,    grho0,  gtolcg, gtolmf, huge,
     +             ierror, info,   ioflag, iout,   iprcnd,
     +             maxit,  mfdopt, mfstat, miu,    mntlcg, mr, 
     +             mrho0,  mxcgit, mxfe,   mxfv,   nbound, nbuk,
     +             oldcos, oldtol, opt,    optgap, optmf , optst,
     +             pcgcos, pcgit,  pcgres, pduit,  pf,     pobj,   
     +             pof,    prttyp, rho0,   root,   s1fctr, stptol, 
     +             stpval, tol1,   tol2,   tolcg,  tolcg0, tolpcg, 
     +             tolslk, tolsw1, tolsw2, sw1, sw2, zero )

c     ------------------------------------------------------------------
c     Print output report header
c     ------------------------------------------------------------------

      call pdphdr(dinfo, info, na, nn, nwkd, maxa, maxn, maxwkd)

c     ------------------------------------------------------------------
c     Check correctness of input data and exit if error is detected.
c     ------------------------------------------------------------------
      call pdchkp(endn, ierror, maxa, maxn, maxwkd, na, nn, nwkd,
     +            prttyp)
     

      if (ierror .ne. 0) then
           call pdsinf( bound,  cgstop, dinfo,  dobj,   dof,    factor,
     +                  fcttol, gap,    grho0,  gtolcg, gtolmf, huge,
     +                  ierror, info,   ioflag, iout,   iprcnd,
     +                  maxit,  mfdopt, mfstat, miu,    mntlcg, mr, 
     +                  mrho0,  mxcgit, mxfe,   mxfv,   nbound, nbuk,
     +                  oldcos, oldtol, opt,    optgap, optmf , optst ,
     +                  pcgcos, pcgit,  pcgres, pduit,  pf,     pobj,   
     +                  pof,    prttyp, rho0,   root,   s1fctr, stptol, 
     +                  stpval, tol1,   tol2,   tolcg,  tolcg0, tolpcg, 
     +                  tolslk, tolsw1, tolsw2, sw1, sw2, zero )
          call pdierr(ierror)
          pdnet = ierror
          return
      end if


c     ------------------------------------------------------------------
c     Initialize maxflow status to inactive.
c     ------------------------------------------------------------------
      mfstat = 0

c     ------------------------------------------------------------------
c     Variable ioflag controls the level of output of summary.
c              ioflag=1: only number of arcs and nodes.
c              ioflag=0: full output.
c     ------------------------------------------------------------------
      ioflag = 0

c     ------------------------------------------------------------------
c     Construct additional data structures.
c     ------------------------------------------------------------------
      call pddstr ( bound,          ptrwki(pibsts), endn,
     +              ptrwki(pilist), ptrwki(piifst), ptrwki(piinxt),
     +              na,             nbound,         nn,
     +              npfrst,         ptrwki(pilost), ptrwki(piofst),
     +              ptrwki(pionxt), strn,           u,
     +              upfrst,         ptrwki(piunxt) )

c     ------------------------------------------------------------------
c     Transform the problem by shifting the lower bounds to zero.
c     ------------------------------------------------------------------
      call pdtrfm( b,              bound,          c,
     +             endn,           fadd,           l,
     +             na,             nbound,         nn,
     +             npfrst,         strn,           u,
     +             upfrst,         ptrwki(piunxt) )

c     ------------------------------------------------------------------
c     Introduce perturbations and transform the data to double 
c     precision.
c     ------------------------------------------------------------------
      call pdprtb( b,              bound,          c,
     +             ptrwkd(prdb00), ptrwkd(prdc00), ptrwkd(prdu00),
     +             endn,           na,             nn,
     +             ptrwkd(prvecs), strn,           u,
     +             upfrst,         ptrwki(piunxt), ptrwkd(prvecw),
     +             ptrwkd(prvecx), ptrwkd(prvecz) )

c     ------------------------------------------------------------------
c     Find the problem data characteristics.
c     ------------------------------------------------------------------
      call pdpdat( ptrwkd(prdb00), bound,          ptrwkd(prdc00),
     +             huge,           maxb,           maxc,
     +             maxu,           minb,           minc, 
     +             minu,           na,             nn,
     +             norb,           norc,           noru,
     +             ptrwkd(prdu00), upfrst,         ptrwki(piunxt) )

c     ------------------------------------------------------------------
c     Check problem feasibility
c     ------------------------------------------------------------------
      call pdckfs(ptrwkd(prdb00), ptrwkd(prdu00), endn, ierror, na, nn, 
     +             strn, ptrwki(piwke1))

      if (ierror .ne. 0) then
          call pdsinf( bound, cgstop, dinfo,  dobj,   dof,  factor,
     +                fcttol, gap,    grho0,  gtolcg, gtolmf, huge,
     +                ierror, info,   ioflag, iout,   iprcnd,
     +                maxit,  mfdopt, mfstat, miu,    mntlcg, mr,
     +                mrho0,  mxcgit, mxfe,   mxfv,   nbound, nbuk,
     +                oldcos, oldtol, opt,    optgap, optmf , optst ,
     +                pcgcos, pcgit,  pcgres, pduit,  pf,     pobj,   
     +                pof,    prttyp, rho0,   root,   s1fctr, stptol, 
     +                stpval, tol1,   tol2,   tolcg,  tolcg0, tolpcg, 
     +                tolslk, tolsw1, tolsw2, sw1, sw2, zero)
          call pdierr(ierror)
          pdnet = ierror
          return
      end if

c     ------------------------------------------------------------------
c     Define the starting values.
c     ------------------------------------------------------------------
      call pdstpt( ptrwkd(prav01), ptrwkd(prkey0), bound,
     +             ptrwkd(prdb00), ptrwkd(prdc00), ptrwkd(prdu00),
     +             endn,           gap,            maxb,
     +             maxc,           minb,           miu,
     +             na,             nbound,         nn,
     +             norb,           npfrst,         root,
     +             ptrwkd(prvecs), s1fctr,         strn,
     +             tolpcg,         upfrst,         ptrwki(piunxt),
     +             ptrwkd(prvecw), ptrwkd(prvecx), ptrwkd(prvecy),
     +             ptrwkd(prvecz) )

c     ------------------------------------------------------------------
c     Initialize interior point iteration counter and define label for 
c     main loop.
c     ------------------------------------------------------------------
      pduit = 0
      info(19) = 3
      prttyp = 3
 30   continue

c     ------------------------------------------------------------------
c     Compute the right-hand-side for preconditioned conjugate gradient.
c     ------------------------------------------------------------------
      call pdcrhs( ptrwkd(prav01), ptrwkd(prav03), ptrwkd(prdb00),
     +             bound,          ptrwki(pibpst), ptrwkd(prdc00),
     +             endn,           larcs,          lfirst,
     +             ptrwki(pilsts), miu,            na, 
     +             nbound,         nn,             npfrst,
     +             ptrwkd(prrhs0), root,           ptrwki(pirsts),
     +             ptrwkd(prvecs), strn,           ptrwkd(prteta),
     +             ptrwkd(prdu00), uarcs,          ufirst,
     +             upfrst,         ptrwki(piunxt), ptrwkd(prvecw),
     +             ptrwkd(prvcwd), ptrwkd(prvecx), ptrwkd(prvcxd),
     +             ptrwkd(prvecy), ptrwkd(prvecz), ptrwkd(prvczd) )

c     ------------------------------------------------------------------
c     Identify maximum spanning tree and build data structures.
c     ------------------------------------------------------------------
 40   continue
      if (iprcnd.eq.1) goto 50
      call pdhprm( bfirst,         ptrwki(pibpst), ptrwki(pibckt),
     +             endn,           ptrwki(pifthr), ptrwki(pifrst),
     +             huge,           ptrwki(piifst), ptrwki(piinxt),
     +             ptrwkd(prkey0), larcs,          lfirst,
     +             ptrwki(pilist), ptrwki(pillnk), ptrwki(pilost),
     +             ptrwki(pilsts), ptrwki(pimark), mr,
     +             na,             nn,             ptrwki(piofst),
     +             ptrwki(pionxt), ptrwki(pipred), ptrwki(pirank),
     +             ptrwki(pirlnk), root,           ptrwki(pirsts),
     +             ptrwki(pirtth), strn,           ptrwkd(prteta),
     +             ptrwki(pitrrc), ptrwki(pitrth), uarcs,
     +             ufirst,         weight )

c     ------------------------------------------------------------------
c     Test optimality of the maximum spanning tree.
c     ------------------------------------------------------------------
      call pdoptm( ptrwkd(prvecp), ptrwkd(prav04), b,
     +  	   bound,	   ptrwki(pibsts), c,
     +  	   dof, 	   endn,	   ptrwki(pifthr),
     +  	   ptrwki(pimark), na,  	   nbound,
     +  	   nn,  	   ptrwki(pilist), npfrst,
     +  	   opt, 	   optflo,	   pf,
     +  	   pof, 	   root,	   ptrwki(pirsts),
     +  	   ptrwki(pirtth), strn,	   ptrwki(pitrrc),
     +  	   ptrwki(pitrth), u,		   uarcs,
     +  	   ufirst,	   upfrst,	   ptrwki(piunxt),
     +  	   ptrwki(pilost), ptrwkd(prvecy), ptrwkd(prvczz),
     +  	   zero )
      if ((pf.eq.1).and.(opt.eq.1)) then
c	   goto 70
      endif

c     ------------------------------------------------------------------
c     Test max flow stopping criterion if miu < 1.0d0.
c     ------------------------------------------------------------------
      if (miu.lt.1.d0) then
         call pdchmf( ptrwkd(prav04), ptrwkd(prdu00), ptrwkd(prdc00),
     +                endn,           mfstat,         mxfe,
     +                mxfv,           nbuk,           na,
     +                nn,             mfdopt,         rc, 
     +                ptrwkd(prvecs), ptrwki(pistdg), stptol,
     +                ptrwki(pisptr), stpval,         strn,
     +                ptrwki(pisvtx),
     +                ptrwkd(prdb00), tol2,           tol1,
     +                ptrwki(piw2e1), ptrwkd(prwrk1),
     +                ptrwki(piwke1), ptrwki(piwke2), ptrwki(piwke3),
     +                ptrwki(piwke4), ptrwkd(prvecw), optflo, 
     +                ptrwkd(prvecx), ptrwkd(prvecy), ptrwkd(prvecz) )

c        ---------------------------------------------------------------
c        Update the tolerances used to guess the optimal partition.
c        ---------------------------------------------------------------
         tol1=tol1*gtolmf
         tol2=tol2/gtolmf

c        ---------------------------------------------------------------
c        Exit on optimal maxflow.
c        ---------------------------------------------------------------
         if (rc.eq.0) goto 70
      endif
 50   continue
      if (pduit.ge.1) then

c        --------------------------------------------------------------
c        Compute objective function.
c        --------------------------------------------------------------
         call pdmkob( ptrwkd(prdb00), bound,          ptrwkd(prdc00),
     +                dobj,           fadd,           na,
     +                nn,             pobj,           root,
     +                ptrwkd(prdu00), upfrst,         ptrwki(piunxt),
     +                ptrwkd(prvecw), ptrwkd(prvecx), ptrwkd(prvecy) )

c        --------------------------------------------------------------
c        Print iteration summary.
c        --------------------------------------------------------------
         call pdsout( cgstop, oldcos, dobj,   dof,    gap,    iout,
     +                iprcnd, pduit,  mfdopt, mfstat, miu,    mxfe,
     +                mxfv,   na,     nn,     opt,    pcgcos, pcgit,
     +                pcgres, oldtol, pf,     pobj,   pof,    prttyp,
     +                tol1,   tol2,   optflo )
      endif

c     ------------------------------------------------------------------
c     Preconditioner is determined by the value of iprcnd.
c
c          iprcnd=1 : Diagonal preconditioner.
c          iprcnd=2 : IQRD preconditioner.
c          iprcnd=3 : Spanning tree preconditioner.
c          iprcnd=4 : Diagonally compensated spanning tree
c                      preconditioner.
c
c     Compute preconditioner.
c     ------------------------------------------------------------------
      if (iprcnd.eq.3) goto 60
      call pdiqrd( ptrwkd(prav04), ptrwkd(prdiag), endn,
     +             ptrwki(pifthr), fcttol,         iprcnd,
     +             na,             ptrwkd(prndig), nn,
     +             rho0,           root,           ptrwki(pirtth),
     +             strn,           ptrwkd(prteta), ptrwki(pitrrc) )
 60   continue

c     ------------------------------------------------------------------
c     Compute the dual direction yd and the gap.
c     ------------------------------------------------------------------
      if (pduit .gt. 1) then
         rgap=dabs(pobj-dobj)/dabs(pobj)
      else
         rgap=1
      endif

c     ------------------------------------------------------------------
c     Compute the tolerances for PCG stopping criterion.
c     ------------------------------------------------------------------
      if (rgap .gt. 1d1) then
         tolcg=tolcg0
      else
         tolcg=tolcg0*rgap
      endif

c     ------------------------------------------------------------------
c     Decide which preconditioner to use.
c     ------------------------------------------------------------------
      if (pduit. gt. 1 .and. iprcnd .eq. 4 .and.
     +     miu  .lt. 1.0d0) iprcnd = 3
      if (tolcg.lt.mntlcg) tolcg=mntlcg
      rho0=rho0*grho0
      if (rho0.lt.mrho0) rho0=mrho0

c     ------------------------------------------------------------------
c     Solve the system using preconditioned conjugate gradient method.
c     ------------------------------------------------------------------
      call pdpcgd( ptrwkd(prrhs0), cgstop,         ptrwkd(prdiag),
     +             endn,           pcgcos,         ptrwki(pifthr), 
     +             iprcnd,         mxcgit,         na,
     +             ptrwkd(prndig), nn,             ptrwkd(prvecp),
     +             pcgit,          ptrwkd(prvecq), ptrwkd(prvecr),
     +             pcgres,         root,           ptrwki(pirtth),
     +             strn,           sw1,            ptrwkd(prteta),
     +             tolcg,          ptrwki(pitrrc), ptrwki(pitrth),
     +             ptrwkd(prvcyd), ptrwkd(prvczz) )

c     ------------------------------------------------------------------
c     Check if change from diagonal preconditioner is needed.
c     ------------------------------------------------------------------
      if ((iprcnd.eq.1).and.(dble(pcgit).gt.sw1)) then

c        ---------------------------------------------------------------
c        Compute objective function.
c        ---------------------------------------------------------------
         call pdmkob( ptrwkd(prdb00), bound,          ptrwkd(prdc00),
     +                dobj,           fadd,           na,
     +                nn,             pobj,           root,
     +                ptrwkd(prdu00), upfrst,         ptrwki(piunxt),
     +                ptrwkd(prvecw), ptrwkd(prvecx), ptrwkd(prvecy) )

c        ---------------------------------------------------------------
c        Print summary of iteration.
c        ---------------------------------------------------------------
         call pdsout( cgstop, oldcos, dobj,   dof,    gap,    iout,
     +                iprcnd, pduit,  mfdopt, mfstat, miu,    mxfe,
     +                mxfv,   na,     nn,     opt,    pcgcos, pcgit,
     +                pcgres, oldtol, pf,     pobj,   pof,    prttyp,
     +                tol1,   tol2,   optflo )

c        ---------------------------------------------------------------
c        Switch to maximum spanning tree preconditioner.
c        ---------------------------------------------------------------
         iprcnd=3

c        ---------------------------------------------------------------
c        Ignore current iteration using diagonal preconditioner.
c        Redo it using maximum spanning tree preconditioner.
c        ---------------------------------------------------------------
         goto 40

      endif

c     ------------------------------------------------------------------
c     Current iteration accepted. Update primal and dual solutions.
c     ------------------------------------------------------------------
      oldtol=tolpcg
      call pdupsl( ptrwkd(prav01), ptrwkd(prkey0), ptrwkd(prav03),
     +             ptrwkd(prav04), ptrwkd(prav05), ptrwkd(prav06),
     +             ptrwkd(prav07), ptrwkd(prav08), ptrwkd(prdb00), 
     +             bound,          ds,             endn,
     +             factor,         gap,            huge,
     +             miu,            na,             nbound,
     +             nn,             npfrst,         ps,
     +             root,           ptrwkd(prvecs), ptrwkd(prvcsd),
     +             strn,           ptrwkd(prteta), tolpcg,
     +             ptrwkd(prdu00), upfrst,         ptrwki(piunxt),
     +             ptrwkd(prvecw), ptrwkd(prvcwd), ptrwkd(prvecx),
     +             ptrwkd(prvcxd), ptrwkd(prvecy), ptrwkd(prvcyd), 
     +             ptrwkd(prvecz), ptrwkd(prvczd) )

c     ------------------------------------------------------------------
c     Increment iteration counter and save PCG cosine tolerance.
c     ------------------------------------------------------------------
      pduit=pduit+1
      oldcos=tolcg

c     ------------------------------------------------------------------
c     Check iteration counter against maximum number of iterations.
c     ------------------------------------------------------------------
      if (pduit.gt.maxit) then

c        ---------------------------------------------------------------
c        Maximum number of iterations reached.
c        Set error flag and go to synchronization before returning.
c        ---------------------------------------------------------------
         ierror = 6
         goto 80
      endif

c     ------------------------------------------------------------------
c     Maximum number of iterations not reached yet, Choose the
c     preconditioner.
c     ------------------------------------------------------------------
      if (iprcnd .eq. 1 .and. pcgit .gt. sw1
     +                   .and. pcgit .lt. sw2) then
          iprcnd = 3
      elseif (iprcnd .eq. 0 .and. pcgit .lt. sw2) then
          iprcnd = 3
      elseif (pcgit .gt. sw2) then
          iprcnd = 0
      endif

c     ------------------------------------------------------------------
c     End of primal-dual iterative loop.
c     ------------------------------------------------------------------
      goto 30
 70   continue

c     ------------------------------------------------------------------
c     Compute objective value at current iteration.
c     ------------------------------------------------------------------
      call pdmkob( ptrwkd(prdb00), bound,          ptrwkd(prdc00),
     +             dobj,           fadd,           na,
     +             nn,             pobj,           root,
     +             ptrwkd(prdu00), upfrst,         ptrwki(piunxt),
     +             ptrwkd(prvecw), ptrwkd(prvecx), ptrwkd(prvecy) )

c     -----------------------------------------------------------------
c     Synchronization: set all the info values in vector info.
c     -----------------------------------------------------------------
 80   call pdsinf( bound,  cgstop, dinfo, dobj,   dof, factor,
     +            fcttol, gap,    grho0,  gtolcg, gtolmf, huge,
     +            ierror, info,   ioflag, iout,   iprcnd,
     +            maxit,  mfdopt, mfstat, miu,    mntlcg, mr, 
     +            mrho0,  mxcgit, mxfe,   mxfv,   nbound, nbuk,
     +            oldcos, oldtol,  opt,  optgap, optmf, optst,
     +            pcgcos, pcgit, pcgres,   pduit,    pf,   pobj,   
     +            pof,  prttyp,   rho0,   root,   s1fctr, stptol, 
     +            stpval, tol1, tol2,   tolcg,  tolcg0, tolpcg, 
     +            tolslk, tolsw1, tolsw2, sw1, sw2, zero )
      call pdierr(ierror)
      pdnet = ierror
      return

c     ------------------------------------------------------------------
c     End of subroutine pdnet.
c     ------------------------------------------------------------------
      end
