






      subroutine pdoptm( alfa,   av4,    b,      bound,  bstats, c, 
     +                   dof,    endn,   father, mark,   na,     nbound,
     +                   nn,     nnodes, npfrst, opt,    optflo, pf,
     +                   pof,    root,   rsets,  rtreth, strn,   trearc,
     +                   treath, u,      uarcs,  ufirst, upfrst, upnext,
     +                   x,      y,      y0,     zero )
c     ------------------------------------------------------------------
c     pdoptm: This subroutine checks the optimality of the sets b, l,
c             and u indicated by the primal-dual interior point method.
c             If the sets are optimal, then the optimal flows are
c             computed.
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             bound : used for computing aditional data structures.
c             na    : number of arcs.
c             nbound: used for computing aditional data structures.
c             nn    : number of nodes.
c             npfrst: auxiliary variable.
c             opt   : flags optimal flow.
c             pf    : iteration level.
c             root  : sets beginning of structure.
c             uarcs : number of arcs at the upper bound.
c             ufirst: auxiliary variable.
c             upfrst: arc with upper bound.
c     ------------------------------------------------------------------
      integer bound , na    , nbound, nn    , npfrst, opt   , pf    ,
     +        root  , uarcs , ufirst, upfrst

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c             b     (1:nn): array of capacities at each node.
c             bstats(1:na): flags if each arc has an upper bound or not.
c             c     (1:na): array of costs at each arc.
c             endn  (1:na): pointer of network data structure.
c             father(1:nn): marks the father of each node.
c             mark  (1:nn): selects nodes to belong to the tree.
c             nnodes(1:nn): number of nodes adjacent to node.
c             optflo(1:na): array with optimal flow at each arc.
c             rsets (1:na): list of arcs at upper bound.
c             rtreth(1:nn): adjacent to a given node.
c             strn  (1:na): adjacency list of network data structure.
c             trearc(1:nn): arc of tree corresponding to node.
c             treath(1:nn): node that is adjacent to given.
c             u     (1:na): upper bound of flow at each arc.
c             upnext(1:na): next arc with positive upper bound.
c             x     (1:nn): current solution.
c     ------------------------------------------------------------------
      integer
     +        b     (1:nn),
     +        bstats(1:na),
     +        c     (1:na),
     +        endn  (1:na),
     +        father(1:nn),
     +        mark  (1:nn),
     +        nnodes(1:nn),
     +        optflo(1:na),
     +        rsets (1:na),
     +        rtreth(1:nn),
     +        strn  (1:na),
     +        trearc(1:nn),
     +        treath(1:nn),
     +        u     (1:na),
     +        upnext(1:na),
     +        x     (1:nn)

c     ------------------------------------------------------------------
c     Double precision input/output arrays:
c
c             alfa(1:nn): auxiliary vector.
c             av4 (1:na): another auxiliary vector.
c             y   (1:nn): variable y.
c             y0  (1:nn): copy of y.
c     ------------------------------------------------------------------
      double precision
     +                 alfa(1:nn),
     +                 av4 (1:na),
     +                 y   (1:nn),
     +                 y0  (1:nn)

c     ------------------------------------------------------------------
c     Integer variables:
c
c             arc   : arc being processed.
c             enod  : current element on endn.
c             i     : loop counter.
c             mnod  : mark of node.
c             node  : another loop counter.
c             pred  : predecessor (father) of node.
c             snod  : current element on strn.
c     ------------------------------------------------------------------
      integer arc   , enod  , i     , mnod  , node  , pred  , snod

c     ------------------------------------------------------------------
c     Double precision input/output variables:
c
c             dof   : dual objective function value on tree.
c             pof   : primal objective function value on tree.
c     ------------------------------------------------------------------
      double precision dof   , pof

c     ------------------------------------------------------------------
c     Double precision variables:
c
c             zero  : value for zero.
c     ------------------------------------------------------------------
      double precision zero

c     ------------------------------------------------------------------
c     Start of executable portion of subroutine.
c
c     Initializations.
c     ------------------------------------------------------------------
      pf  = 0
      opt = 0
      pof = 0d0
      dof = 0d0

c     ------------------------------------------------------------------
c     Initialize flow variables x.
c     ------------------------------------------------------------------
      do 10 node=1,root-1
         x(node) =  b(node)
 10   continue
      do 20 node=root+1,nn
         x(node) =  b(node)
 20   continue
      x(root) = 0
      arc=ufirst
      do 30 i=1,uarcs
         snod = strn(arc)
         enod = endn(arc)
         x(snod) = x(snod)+u(arc)
         x(enod) = x(enod)-u(arc)
         arc = rsets(arc)
 30   continue
      x(root) = 0

c     ------------------------------------------------------------------
c     Compute flow variables x belonging to the set b by
c     performing a reverse treath traversal of the spanning
c     tree. This procedure computes also the total cost of
c     the basic variables.
c     ------------------------------------------------------------------
      pf = 1
      node = rtreth(root)
      do 40 i=2,nn
         arc  = trearc(node)
         snod = strn(arc)
         pred = father(node)
         x(pred) = x(pred)+x(node)
         if (node.eq.snod) x(node)=-x(node)
         pof = pof + dble(x(node))*dble(c(arc))
         if (bstats(arc).eq.1) then
            if (x(node).lt.0.or.x(node).gt.u(arc)) then
               pf = 0
               return
            endif
         else
            if (x(node).lt.0) then
               pf = 0
               return
            endif
         endif
         node = rtreth(node)
 40   continue
      x(root) = 0

c     ------------------------------------------------------------------
c     Add the total cost of the flow variables at the upper bound.
c     ------------------------------------------------------------------
      arc=ufirst
      do 50 i=1,uarcs
         snod = strn(arc)
         enod = endn(arc)
         pof = pof + dble(c(arc))*dble(u(arc))
         arc  = rsets(arc)
 50   continue

c     ------------------------------------------------------------------
c     Project y onto the dual face complementary to x.
c     ------------------------------------------------------------------
      y0(root)   = 0d0
      mark(root) = root
      node = treath(root)
      do 60 i=2,nn
         arc             = trearc(node)
         if (x(node).eq.0.or.(bstats(arc).eq.1.
     +        and.x(node).eq.u(arc))) then
            y0(node)     = 0d0
            mark(node)   = node
            nnodes(node) = 1
            alfa(node)   = 0d0
         else
            pred         = father(node)
            arc          = trearc(node)
            y0(node)     = dble(c(arc))
            mark(node)   = mark(pred)
            mnod         = mark(node)
            nnodes(mnod) = nnodes(mnod)+1
         endif
         node = treath(node)
 60   continue
      node = treath(root)
      do 70 i=2,nn
         arc  = trearc(node)
         snod = strn(arc)
         enod = endn(arc)
         mnod = mark(node)
         if ( (bstats(arc).eq.0.and.x(node).gt.0).or.
     +        (bstats(arc).eq.1.and.x(node).gt.0.
     +        and.x(node).lt.u(arc))) then
            if (node.eq.enod) then
               y0(enod) = y0(enod)+y0(snod)
            else
               y0(snod) = y0(enod)-y0(snod)
            endif
         endif
         alfa(mnod) = alfa(mnod)+y(node)-y0(node)
         node = treath(node)
 70   continue

      do 80 node=1,root-1
         mnod = mark(node)
         y0(node) = y0(node)+alfa(mnod)/dble(nnodes(mnod))
 80   continue
      do 90 node=root+1,nn
         mnod = mark(node)
         y0(node) = y0(node)+alfa(mnod)/dble(nnodes(mnod))
 90   continue

c     ------------------------------------------------------------------
c     Estimate dual objective function value.
c     ------------------------------------------------------------------
      do 100 arc=1,na
         snod = strn(arc)
         enod = endn(arc)
         av4(arc) = y0(enod)-y0(snod)-dble(c(arc))
 100  continue

      arc = npfrst
      do 110 i=1,nbound
         if (av4(arc).gt.zero) return
         arc = upnext(arc)
 110  continue

      do 120 node=1,root-1
         dof = dof+dble(b(node))*y0(node)
 120  continue
      do 130 node=root+1,nn
         dof = dof+dble(b(node))*y0(node)
 130  continue

      arc = upfrst
      do 140 i=1,bound
         snod = strn(arc)
         enod = endn(arc)
         dof  = dof-dble(u(arc))*dmax1(av4(arc),0.0d0)
         arc  = upnext(arc)
 140  continue
      if (dabs(pof-dof).lt.1.) then
c        ----------------------------------------------------------------
c        Dump the optimal flow to optflo and exit.
c        ----------------------------------------------------------------
         opt = 1
         do 150 i = 1, na
            optflo(i) = 0
 150     continue
         do 160 i = 1, nn
            optflo(trearc(i)) = x(i)
 160     continue
         return
      endif

c     ------------------------------------------------------------------
c     End of subroutine pdoptm.
c     ------------------------------------------------------------------
      end
