      subroutine pdpcgd( b,      cgstop, diag,   endn,   err,    father,
     +                   iprcnd, mxcgit, na,     ndiag,  nn,     p, 
     +                   pcgit,  q,      r,      res,    root,   rtreth, 
     +                   strn,   sw1,    teta,   tolcg,  trearc, treath, 
     +                   x,      z )
c     ------------------------------------------------------------------
c     pdpcgd: This sobroutine solves the system a*teta*at x = b by means
c             of a preconditioned conjugate gradient (pcg) algorithm.
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             cgstop: flags the way PCG stopped.
c             iprcnd: preconditioner used.
c             mxcgit: maximum number of PCG iterations.
c             na    : number of arcs.
c             nn    : number of nodes.
c             pcgit : number of PCG iterations performed.
c             root  : sets beginning of structure.
c     ------------------------------------------------------------------
      integer cgstop, iprcnd, mxcgit, na    , nn    , pcgit  , root

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c             endn  (1:na): pointer of network data structure.
c             father(1:nn): marks the father of each node.
c             rtreth(1:nn): adjacent to a given node.
c             strn  (1:na): adjacency list of network data structure.
c             trearc(1:nn): arc of tree corresponding to node.
c             treath(1:nn): node that is adjacent to given.
c     ------------------------------------------------------------------
      integer
     +        endn  (1:na),
     +        father(1:nn),
     +        rtreth(1:nn),
     +        strn  (1:na),
     +        trearc(1:nn),
     +        treath(1:nn)

c     ------------------------------------------------------------------
c     Double precision input/output arrays:
c
c             b     (1:nn): array of capacities at each node.
c             diag  (1:nn): diagonal of matrix.
c             ndiag (1:nn): diagonal of preconditioner.
c             p     (1:nn): vector p.
c             q     (1:nn): vector q.
c             r     (1:nn): vector r.
c             teta  (1:na): diagonal theta.
c             x     (1:nn): primal solution x.
c             z     (1:nn): dual solution z.
c     ------------------------------------------------------------------
      double precision
     +                  b     (1:nn),
     +                  diag  (1:nn),
     +                  ndiag (1:nn),
     +                  p     (1:nn),
     +                  q     (1:nn),
     +                  r     (1:nn),
     +                  teta  (1:na),
     +                  x     (1:nn),
     +                  z     (1:nn)

c     ------------------------------------------------------------------
c     Double precision input/output variables:
c
c             err   : value of cosine on PCG iteration.
c             res   : value of residual on PCG iteration.
c             tolcg : tolerance for PCG stopping criterion.
c     ------------------------------------------------------------------
      double precision err   , res   , tolcg

c     ------------------------------------------------------------------
c     Double precision variables:
c
c             alfa  : parameter from PCG algorithm.
c             beta  : parameter from PCG algorithm.
c             dnrm2 : 2-norm of a vector (BLAS function).
c             eps   : temporary storage for inner product.
c             nrmb  : 2-norm of vector b.
c             oldeps: value of eps in previous iteration.
c             sw1   : copy of maximum iteration count.
c     ------------------------------------------------------------------
      double precision alfa  , beta  , dnrm2 , eps   , nrmb  , oldeps,
     +                 sw1

c     ------------------------------------------------------------------
c     Start of executable section of subroutine.
c
c     Initializations.
c     ------------------------------------------------------------------
c
      pcgit = 0
      call pdattx( endn,   na,     nn,     root,   strn,   teta,
     +             x,      r )
      call dcopy( nn,     r,      1,      z,      1 )
      call pdsbvc( nn,     root,   b,      z,      r )
      call pdsviq( r,      diag,   father, iprcnd, na,     ndiag,
     +             nn,     root,   rtreth, teta,   trearc, treath,
     +             z )
      call pdsteq( nn,     root,   p,      z )
      call pdpryx( eps,    nn,     root,   r,      z )
      call pdresd( res,    nn,     r )

c     ------------------------------------------------------------------
c     PCG algorithm.
c     ------------------------------------------------------------------
      nrmb= dnrm2( nn,     b,      1 )
 10   continue
      pcgit = pcgit + 1
      if (pcgit.ge.mxcgit) then
         cgstop=0
         return
      endif
      if ((iprcnd.eq.1).and.(dble(pcgit).gt.sw1)) then
         cgstop=1
         return
      endif
      call pdattx( endn,   na,     nn,     root,   strn,   teta,
     +             p,      q )
      call pdpryx( alfa,   nn,     root,   p,      q )
      alfa = eps/alfa
      call pdupv1( alfa,   nn,     root,   x,      p )
      alfa = -alfa
      call pdupv1( alfa,   nn,     root,   r,      q )
      call pdsviq( r,      diag,   father, iprcnd, na,     ndiag,
     +             nn,     root,   rtreth, teta,   trearc, treath,
     +             z )
      oldeps = eps
      call pdpryx( eps,    nn,     root,   r,      z )
      beta = eps/oldeps
      call pdupv2( beta,   nn,     root,   p,      z )
      call pdresd( res,    nn,     r )
c     ------------------------------------------------------------------
c     Cosine stopping criterion.
c     ------------------------------------------------------------------
      call pdcosn( b,      err,    nn,     nrmb,   q,      r,
     +             root )
      if (err.lt.tolcg) then
         cgstop=2
         return
      endif
      goto 10

c     ------------------------------------------------------------------
c     End of subroutine pdpcgd.
c     ------------------------------------------------------------------
      end
