/*-
 * $Id: pdrdmc.c,v 1.1.1.1 2002/09/25 18:11:29 ggv Exp $
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Contents: Functions to read a DIMACS-format network data file. It 
 *   includes an interface suitable for use with FORTRAN programs.
 *
 * Authors: Luis Portugal (Univ. Coimbra - Portugal)
 *          Mauricio G. C. Resende (AT&T Labs Research - USA)
 *          Geraldo Veiga (AT&T Labs - USA)
 *          Joaquim Judice (Univ. Coimbra - Portugal)
 *          Joao Patricio (EST Tomar - Portugal)
 *
 * User callable functions:
 *
 * >>> pdrdmm <<<
 * >>> pdrdmc_ <<<
 * >>> pdrdmc <<< 
 *
 * Revision History: 
 *   2001-12-27	Geraldo Veiga
 *       Initial release for ACM distribution.
 *
 */

/* Include files */
#include "pdnet.h"		/* general definitions for PDNET ANSI C */
#include "debug.h"             /* PDNET debug macros */

/* Prototypes for fuction in this module */
void pdrdmm (int *nvrtx_ptr,	/* pointer to number of vertices */
	     int *nedge_ptr,	/* pointer to number of edges */
	     int **strv_ptr,	/* pointer to start vertex array */
	     int **endv_ptr,	/* pointer to end vertex array */
	     int **supply_ptr,	/* pointer to node flow supply array */
	     int **cost_ptr,	/* pointer to cost coefficient array */
	     int **lwbnd_ptr,	/* pointer to lower bound array */
	     int **cap_ptr	/* pointer to edge capacity array */
  );
  
void pdrdmc_ (int *nvrtx_ptr,	/* pointer to number of vertices */
	      int *nedge_ptr,	/* pointer to number of edges */
	      int *strv,	/* start vertex array */
	      int *endv,	/* end vertex array */
	      int *supply,	/* node flow supply array */
	      int *cost,	/* cost coefficient array */
	      int *lwbnd,	/* lower bound array */
	      int *cap,	/* edge capacity array */
	      int *max_vrtx_ptr,	/* pointer to maximum number of vertices */
	      int *max_edge_ptr	/* pointer to maximum number of edges */
  );

void pdrdmc (int *nvrtx_ptr,	/* pointer to number of vertices */
	     int *nedge_ptr,	/* pointer to number of edges */
	     int **strv_ptr,	/* pointer to start vertex array */
	     int **endv_ptr,	/* pointer to end vertex array */
	     int **supply_ptr,	/* pointer to node flow supply array */
	     int **cost_ptr,	/* pointer to cost coefficient array */
	     int **lwbnd_ptr,	/* pointer to lower bound array */
	     int **cap_ptr,	/* pointer to edge capacity array */
	     int max_vrtx,	/* maximum number of vertices */
	     int max_edge	/* maximum number of edges */
  );
  
 /*-
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * >>> pdrdmm <<<
 *
 * Description: FORTRAN interface for function pdrdmc.  This interface
 *  does not allow for dynamic allocation of the network data structure.
 *
 * Return Value: none
 *
 * Side effects: Returns network data structure via function arguments.
 *   Data structure arrays are dynamically allocated .
 * --------------------------------------------------------------------
 */

void
pdrdmm (int *nvrtx_ptr,	/* pointer to number of vertices */
	int *nedge_ptr,	/* pointer to number of edges */
	int **strv_ptr,	/* pointer to start vertex array */
	int **endv_ptr,	/* pointer to end vertex array */
	int **supply_ptr,	/* pointer to node flow supply array */
	int **cost_ptr,	/* pointer to cost coefficient array */
	int **lwbnd_ptr,	/* pointer to lower bound array */
	int **cap_ptr		/* pointer to edge capacity array */
  )
{

  (void) pdrdmc (nvrtx_ptr, nedge_ptr, strv_ptr, endv_ptr, supply_ptr, 
                 cost_ptr, lwbnd_ptr, cap_ptr, 0, 0);
}

 /*-
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * >>> pdrdmc_ <<<
 *
 * Description: FORTRAN interface for function pdrdmc.  This interface
 *  does not allow for dynamic allocation of the network data structure.
 *
 * Return Value: none
 *
 * Side effects: Returns network data structure via function arguments.
 * --------------------------------------------------------------------
 */

void
pdrdmc_ (int *nvrtx_ptr,	/* pointer to number of vertices */
	 int *nedge_ptr,	/* pointer to number of edges */
	 int *strv,		/* start vertex array */
	 int *endv,		/* end vertex array */
	 int *supply,		/* node flow supply array */
	 int *cost,		/* cost coefficient array */
	 int *lwbnd,		/* lower bound array */
	 int *cap,		/* edge capacity array */
	 int *max_vrtx_ptr,	/* pointer to maximum number of vertices */
	 int *max_edge_ptr	/* pointer to maximum number of edges */
  )
{

  (void) pdrdmc (nvrtx_ptr, nedge_ptr, &strv, &endv, &supply, &cost,
		 &lwbnd, &cap, *max_vrtx_ptr, *max_edge_ptr);
}

/*-
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * >>> pdrdmc <<<
 *
 * Description: Reads a DIMACS-format network data file, creating the
 *   data structure returned via function arguments. When invoked with 
 *   max_vrtx = 0 and max_edge = 0, this function also allocates 
 *   memory for the network data structure.
 * 
 * Return Value: none
 *
 * Side effects: Returns network data structure via function arguments.
 *   Data structure arrays are dynamically allocated when requested via
 *   function arguments.
 * --------------------------------------------------------------------
 */

void
pdrdmc (int *nvrtx_ptr,	/* pointer to number of vertices */
	int *nedge_ptr,	/* pointer to number of edges */
	int **strv_ptr,	/* pointer to start vertex array */
	int **endv_ptr,	/* pointer to end vertex array */
	int **supply_ptr,	/* pointer to node flow supply array */
	int **cost_ptr,	/* pointer to cost coefficient array */
	int **lwbnd_ptr,	/* pointer to lower bound array */
	int **cap_ptr,		/* pointer to edge capacity array */
	int max_vrtx,		/* maximum number of vertices */
	int max_edge		/* maximum number of edges */
  )
{
  int nvrtx,			/* number of vertices */
    nedge,			/* number of edges */
   *strv,			/* start vertex array */
   *endv,			/* end vertex array */
   *supply,			/* node flow supply array */
   *cost,			/* cost coefficient array */
   *lwbnd,			/* lower bound array */
   *cap;			/* edge capacity array */

  FILE *file;			/* input file */

  char ctmp,			/* temporary char */
    in_fmt[4];			/* input file format = min, max, asn */

  int cap_sum_s = 0,		/* sum of capacities from source */
    cap_sum_t = 0,		/* sum of capacities into sink */
    itmp,			/* temporary  integer */
    sink = 0,			/* sink vertex in max flow */
    source = 0,			/* source vertex in max flow */
    v1,				/* vertex index */
    v2,				/* vertex index */
    vrtx;			/* vertex index */

  register int lines = 0,	/* number of lines in file */
    edg = 0;			/* edge counter */

  int command;			/* first char in command line */

  BOOL p_pass = FALSE,		/* prob def line read */
    n_pass = FALSE,		/* node line read */
    a_pass = FALSE;		/* arc line read */

/* Copy argument values to local variables */
  nvrtx = *nvrtx_ptr;
  nedge = *nedge_ptr;
  strv = *strv_ptr;
  endv = *endv_ptr;
  supply = *supply_ptr;
  cost = *cost_ptr;
  lwbnd = *lwbnd_ptr;
  cap = *cap_ptr;

/* Set input file to standard input */
  file = stdin;

/* Read input file */
  while ((command = getc (file)) != EOF)
    {
      lines++;			/* increment line counter */

      switch (command)
	{
	case 'c':		/* comment line */

	  if (fscanf (file, "%*[^\n]\n") == 0)
	    fscanf (file, "\n");
	  break;

	case 'p':		/* prob def line */

	  if (p_pass)
	    usrerr ("Repeated problem descriptor - line %d", lines);
	  else
	    p_pass = TRUE;

	  fscanf (file, "%3s %d %d\n", in_fmt, &nvrtx, &nedge);

	  /* Allocate arrays */
	  if (max_vrtx == (int) 0)
	    {
	      MALLOC (supply, int, nvrtx);
	      max_vrtx = nvrtx;
	    }
	  if (max_edge == (int) 0)
	    {
	      MALLOC (strv, int, nedge);
	      MALLOC (endv, int, nedge);
	      MALLOC (cost, int, nedge);
	      MALLOC (cap, int, nedge);
	      MALLOC (lwbnd, int, nedge);
	      max_edge = nedge;
	    }

	  /* Test configuration */
	  if (nvrtx > max_vrtx)
	    usrerr ("Number of vertices(%d) exceed maximum configuration(%d)",
		    nvrtx, max_vrtx);

	  if (nedge > max_edge)
	    usrerr ("Number of edges(%d) exceed maximum configuration(%d)",
		    nedge, max_edge);

	  /* Initialize arrays */
	  for (vrtx = 0; vrtx < nvrtx; vrtx++)
	    supply[vrtx] = (int) 0;

	  /* Adjust number of edges  for max flow */
	  if (strncmp (in_fmt, "max", 3) == 0)
	    nedge++;

	  /* Initialize supply for assignment (all sinks) */
	  if (strncmp (in_fmt, "asn", 3) == 0)
	    for (vrtx = 0; vrtx < nvrtx; vrtx++)
	      supply[vrtx] = (int) 1.0;

	  break;

	case 'n':		/* node line */

	  /* Check consistensy */
	  if (!p_pass)
	    usrerr ("Missing problem descriptor - line %d", lines);
	  if (a_pass)
	    usrerr ("Node descriptor out of sequence - line %d", lines);
	  n_pass = TRUE;

	  /* Scan input line */
	  if (strncmp (in_fmt, "min", 3) == 0)
	    {

	      /* Minimum-cost flow */
	      fscanf (file, "%d", &vrtx);
	      fscanf (file, "%d\n", &itmp);
	      supply[vrtx - 1] = (int) -itmp;
	    }
	  else if (strncmp (in_fmt, "max", 3) == 0)
	    {

	      /* Maximum flow */
	      fscanf (file, "%d", &vrtx);
	      fscanf (file, "%s\n", &ctmp);
	      switch (ctmp)
		{
		case 's':
		  if (source != 0)
		    usrerr ("Repeated source node - line %d", lines);
		  source = vrtx;
		  break;

		case 't':
		  if (sink != 0)
		    usrerr ("Repeated sink node - line %d", lines);
		  sink = vrtx;
		  break;

		default:
		  usrerr ("Node descriptor not recognized - line %d", lines);
		  break;
		}
	    }
	  else if (strncmp (in_fmt, "asn", 3) == 0)
	    {

	      /* Assignment */
	      fscanf (file, "%d\n", &vrtx);
	      supply[vrtx - 1] = (int) -1.0;
	    }
	  break;

	case 'a':		/* arc line */

	  /* Check consistensy */
	  if (!p_pass)
	    usrerr ("Missing problem descriptor - line %d", lines);
	  a_pass = TRUE;

	  /* Scan input line */
	  if (++edg > nedge)
	    usrerr ("Too many edge descriptors - line %d", lines);

	  if (strncmp (in_fmt, "min", 3) == 0)
	    {

	      /* Minimum-cost flow */
	      fscanf (file, "%d ", &itmp);
	      if ((itmp > nvrtx) || (itmp < (int) 1)) {
	        DB_PRINT_2(0, "itmp nvrtx: %ld:%ld" , itmp, nvrtx);
		usrerr ("Vertex index out of range - line %d", lines);
	      }
	      strv[edg - 1] = (int) itmp;

	      fscanf (file, "%d ", &itmp);
	      if ((itmp > nvrtx) || (itmp < (int) 1)) {
	        DB_PRINT_2(0, "itmp nvrtx: %ld:%ld" , itmp, nvrtx);
		usrerr ("Vertex index out of range - line %d", lines);
	      }
	      endv[edg - 1] = (int) itmp;

	      fscanf (file, "%d", &itmp);
	      lwbnd[edg - 1] = (int) itmp;

	      fscanf (file, "%d", &itmp);
	      cap[edg - 1] = (int) itmp;

	      fscanf (file, "%d\n", &itmp);
	      cost[edg - 1] = (int) itmp;
	    }
	  else if (strncmp (in_fmt, "max", 3) == 0)
	    {

	      /* Maximum flow */
	      fscanf (file, "%d", &v1);
	      if (v1 > nvrtx || v1 < 1)
		usrerr ("Vertex index out of range - line %d", lines);
	      strv[edg - 1] = (int) v1;

	      fscanf (file, "%d", &v2);
	      if (v2 > nvrtx || v2 < 1)
		usrerr ("Vertex index out of range - line %d", lines);
	      endv[edg - 1] = (int) v2;

	      fscanf (file, "%d\n", &itmp);
	      cap[edg - 1] = (int) itmp;
	      if (v1 == source)
		cap_sum_s += itmp;
	      if (v2 == sink)
		cap_sum_t += itmp;
	      cost[edg - 1] = (int) 0.;
	      lwbnd[edg - 1] = (int) 0.;
	    }
	  else if (strncmp (in_fmt, "asn", 3) == 0)
	    {

	      /* Assignment */
	      fscanf (file, "%d", &itmp);
	      if (itmp > nvrtx || itmp < 1)
		usrerr ("Vertex index out of range - line %d", lines);
	      strv[edg - 1] = (int) itmp;
	      fscanf (file, "%d", &itmp);
	      if (itmp > nvrtx || itmp < 1)
		usrerr ("Vertex index out of range - line %d", lines);
	      endv[edg - 1] = (int) itmp;
	      fscanf (file, "%d\n", &itmp);
	      cost[edg - 1] = (int) itmp;
	      cap[edg - 1] = (int) 1.0;
	      lwbnd[edg - 1] = (int) 0.0;
	    }
	  break;

	default:		/* error */
	  usrerr ("Command not recognized - line %d", lines);
	  break;

	}
    }

/* Add circulation edge for max flow problem */
  if (strncmp (in_fmt, "max", 3) == 0)
    {
      if (source + sink == 0)
	usrerr ("Missing source or sink node descriptor - line %d", lines);
      strv[nedge - 1] = sink;
      endv[nedge - 1] = source;
      cost[nedge - 1] = (int) -1.;
      cap[nedge - 1] = (int) MIN (cap_sum_s, cap_sum_t);
      lwbnd[nedge - 1] = (int) 0.0;
    }

/* Copy local variables values to function arguments */
  *nvrtx_ptr = nvrtx;
  *nedge_ptr = nedge;
  *strv_ptr = strv;
  *endv_ptr = endv;
  *supply_ptr = supply;
  *cost_ptr = cost;
  *lwbnd_ptr = lwbnd;
  *cap_ptr = cap;
}
