






      subroutine pdstpt( av1,    av2,    bound , db,     dc,     du,
     +                   endn,   gap,    maxb,   maxc,   minb,   miu,   
     +                   na,     nbound, nn,     norb,   npfrst, root, 
     +                   s,      s1fctr, strn,   tolpcg, upfrst, upnext, 
     +                   w,      x,      y,      z )
c     ------------------------------------------------------------------
c     pdstpt: This subroutine computes the starting values for the
c             primal-dual interior point algorithm.
c     ------------------------------------------------------------------
      implicit none

c     ------------------------------------------------------------------
c     Integer input/output variables:
c
c             bound : number of bounded arcs.
c             na    : number of arcs.
c             nbound: number of unbounded arcs.
c             nn    : number of nodes.
c             npfrst: next bounded arc to be considered.
c             root  : sets beginning of structure.
c             upfrst: next unbounded arc to be considered.
c     ------------------------------------------------------------------
      integer bound , na    , nbound, nn    , npfrst, root  ,  upfrst

c     ------------------------------------------------------------------
c     Integer input/output arrays:
c
c          endn  (1:na): pointer of network data structure.
c          strn  (1:na): adjacency list of network data structure.
c          upnext(1:na): next arc with positive upper bound.
c     ------------------------------------------------------------------
      integer
     +     endn  (1:na),
     +     strn  (1:na),
     +     upnext(1:na)

c     ------------------------------------------------------------------
c     Double precision input/output arrays:
c
c           av1(1:na): temporary array 1.
c           av2(1:nn): temporary array 2.
c           db (1:nn): array of capacities at each node.
c           dc (1:na): array of costs at each arc.
c           du (1:na): upper bound of flow at each arc.
c           s  (1:na): dual iterate s.
c           w  (1:na): dual iterate w.
c           x  (1:na): primal iterate x.
c           y  (1:nn): dual iterate y.
c           z  (1:na): primal iterate z.
c     ------------------------------------------------------------------
      double precision
     +      av1(1:na),
     +      av2(1:nn),
     +      db (1:nn),
     +      dc (1:na),
     +      du (1:na),
     +      s  (1:na),
     +      w  (1:na),
     +      x  (1:na),
     +      y  (1:nn),
     +      z  (1:na)

c     ------------------------------------------------------------------
c     Integer variables:
c
c             arc   : current arc.
c             enod  : current element on endn.
c             i     : loop counter.
c             node  : another loop counter.
c             snod  : current element on strn.
c     ------------------------------------------------------------------
      integer arc   , enod  , i     , node  , snod

c     ------------------------------------------------------------------
c     Double precision input/output variables:
c
c             gap   : value of the gap.
c             maxb  : largest absolute value of b.
c             maxc  : largest absolute value of c.
c             minb  : smallest absolute value of b.
c             miu   : value of interior-point mu parameter.
c             norb  : 1-norm of vector b.
c             s1fctr: factor for miu.
c             tolpcg: initial value for tolcg.
c     ------------------------------------------------------------------
      double precision gap   , maxb  , maxc  , minb  , miu   , norb  ,
     +                 s1fctr, tolpcg

c     ------------------------------------------------------------------
c     Double precision variables:
c
c              adema : initial demand.
c              aux1  : auxiliary variable 1.
c              aux2  : auxiliary variable 2.
c              demand: demand.
c              nav1  : small quantity.
c     ------------------------------------------------------------------
      double precision adema , aux1  , aux2  , demand, nav1

c     ------------------------------------------------------------------
c     External functions:
c
c             dnrm2  : euclidean norm of a vector (BLAS).
c     ------------------------------------------------------------------
      double precision    dnrm2

c     ------------------------------------------------------------------
c     Start of executable section of subroutine.
c     ------------------------------------------------------------------
      adema  = dmax1(dabs(minb),dabs(maxb))
      demand = norb/2d0

c     ------------------------------------------------------------------
c     Computing auxilliary vector av2.
c     ------------------------------------------------------------------
      do 10 node=1,nn
         av2(node) = db(node)
 10   continue

c     ------------------------------------------------------------------
c     Computing initial vector y.
c     ------------------------------------------------------------------
      aux1 = maxc/adema
      do 20 node=1,root-1
         y(node) = aux1*db(node)
 20   continue
      do 30 node=root+1,nn
         y(node) = aux1*db(node)
 30   continue
      y(root) = 0d0

c     ------------------------------------------------------------------
c     Computing auxilliary vector av1.
c     ------------------------------------------------------------------
      do 40 arc=1,na
         snod = strn(arc)
         enod = endn(arc)
         av1(arc) = y(snod)-y(enod)+dc(arc)
 40   continue

c     ------------------------------------------------------------------
c     Computing initial miu.
c     ------------------------------------------------------------------
      miu      = 0d0
      arc      = npfrst
      do 50 i=1,nbound
         aux1  = dabs(av1(arc))
         miu   = dmax1(miu,aux1)
         arc   = upnext(arc)
 50   continue
      miu      = demand*miu
      arc      = upfrst
      do 60 i=1,bound
         aux1  = dabs(av1(arc)*du(arc))
         miu   = dmax1(miu,aux1)
         arc   = upnext(arc)
 60   continue
      miu = s1fctr*miu

c     ------------------------------------------------------------------
c     Computing initial values for variables with upper bound.
c     ------------------------------------------------------------------
      nav1=1.0d-8 * dnrm2 ( na, av1, 1)
      arc = upfrst
      do 70 i=1,bound
         aux1 = du(arc)/2d0
         if (av1(arc).gt. nav1) then
            aux2   = miu/av1(arc)
            x(arc) = aux1+aux2-dsqrt(aux1*aux1+aux2*aux2)
         else
            if (av1(arc).lt.-nav1) then
               aux2   = miu/av1(arc)
               x(arc) = aux1+aux2+dsqrt(aux1*aux1+aux2*aux2)
            else
               x(arc) = aux1
            endif
         endif
         s(arc)    = du(arc)-x(arc)
         z(arc)    = miu/x(arc)
         w(arc)    = miu/s(arc)
         arc = upnext(arc)
 70   continue

c     ------------------------------------------------------------------
c     Computing initial values for variables without upper bound.
c     ------------------------------------------------------------------
      arc    = npfrst
      do 80 i=1,nbound
         aux1 = demand/2d0
         if (av1(arc).gt.0.) then
            aux2   = miu/av1(arc)
            x(arc) = aux1+aux2-dsqrt(aux1*aux1+aux2*aux2)
         else
            if (av1(arc).lt.0.) then
               aux2   = miu/av1(arc)
               x(arc) = aux1+aux2+dsqrt(aux1*aux1+aux2*aux2)
            else
               x(arc) = aux1
            endif
         endif
         z(arc)  = miu/x(arc)
         arc = upnext(arc)
 80   continue

c     ------------------------------------------------------------------
c     Computing auxialliary vector av2.
c     ------------------------------------------------------------------
       do 90 arc=1,na
         snod = strn(arc)
         enod = endn(arc)
         av2(snod) = av2(snod)+x(arc)
         av2(enod) = av2(enod)-x(arc)
 90   continue
      av2(root) = 0d0

c    ------------------------------------------------------------------
c    Computing initial values to miu and tolpcg.
c    ------------------------------------------------------------------
      tolpcg    = 0d0
      do 100 node=1,nn
         tolpcg = tolpcg+dabs(av2(node))
 100  continue
      gap = miu*dble(na)
      miu =  gap/dble(10*na)
      tolpcg = tolpcg/1d1

c     ------------------------------------------------------------------
c     End of subroutine pdstpt.
c     ------------------------------------------------------------------
      end
